from unittest import TestCase
from uuid import UUID

from pycarlo.common import MCONParser


class MCONParserTests(TestCase):
    _TEST_ACCOUNT_ID = '12345678-1234-5678-1234-567812345678'
    _TEST_RESOURCE_ID = '21345678-1234-5678-1234-567812345678'
    _TEST_TABLE_ID = 'db:sch.tbl'

    def test_valid_table_mcon(self):
        mcon = f'MCON++{self._TEST_ACCOUNT_ID}++{self._TEST_RESOURCE_ID}++table++{self._TEST_TABLE_ID}'
        parsed_mcon = MCONParser.parse_mcon(mcon)
        self.assertIsNotNone(parsed_mcon)
        self.assertEqual(UUID(self._TEST_ACCOUNT_ID), parsed_mcon.account_id)
        self.assertEqual(UUID(self._TEST_RESOURCE_ID), parsed_mcon.resource_id)
        self.assertEqual('table', parsed_mcon.object_type)
        self.assertEqual(self._TEST_TABLE_ID, parsed_mcon.object_id)

    def test_invalid_mcon(self):
        mcon = 'invalid'
        parsed_mcon = MCONParser.parse_mcon(mcon)
        self.assertIsNone(parsed_mcon)

    def test_valid_field_mcon(self):
        field_id = f'{self._TEST_TABLE_ID}+++field_id'
        mcon = f'MCON++{self._TEST_ACCOUNT_ID}++{self._TEST_RESOURCE_ID}++field++{field_id}'
        parsed_mcon = MCONParser.parse_mcon(mcon)
        self.assertIsNotNone(parsed_mcon)
        self.assertEqual(UUID(self._TEST_ACCOUNT_ID), parsed_mcon.account_id)
        self.assertEqual(UUID(self._TEST_RESOURCE_ID), parsed_mcon.resource_id)
        self.assertEqual('field', parsed_mcon.object_type)
        self.assertEqual(field_id, parsed_mcon.object_id)



from sgqlc.operation import Operation

from pycarlo.lib import schema


class Query(Operation):
    def __init__(self, *args, **kwargs):
        """
        An MCD Query operation.

        Supports all Operation params and functionality.
        """
        super().__init__(typ=schema.Query, *args, **kwargs)


class Mutation(Operation):
    def __init__(self, *args, **kwargs):
        """
        An MCD Mutation operation.

        Supports all Operation params and functionality.
        """
        super().__init__(typ=schema.Mutation, *args, **kwargs)



import json
import time
from typing import Optional, Callable, List, Union
from uuid import UUID
from warnings import warn

from box import Box

from pycarlo.common import get_logger
from pycarlo.common.utils import boxify
from pycarlo.core import Client, Mutation, Query
from pycarlo.features.circuit_breakers.exceptions import CircuitBreakerPollException, CircuitBreakerPipelineException
from pycarlo.lib.schema import CircuitBreakerState, SqlJobCheckpointStatus

logger = get_logger(__name__)


class CircuitBreakerService:
    _TERM_STATES = {'PROCESSING_COMPLETE', 'HAS_ERROR'}

    def __init__(self,
                 mc_client: Optional[Client] = None,
                 print_func: Optional[Callable] = logger.info):
        """
        Convenience methods to help with using circuit breaker rules.

        :param mc_client: MCD client (e.g. for creating a custom session); created otherwise.
        :param print_func: Function to use for echoing. Uses python logging by default, which requires setting MCD_VERBOSE_ERRORS.
        """
        self._client = mc_client or Client()
        self._print_func = print_func

    def trigger_and_poll(self, rule_uuid: Optional[Union[str, UUID]] = None, namespace: Optional[str] = None,
                         rule_name: Optional[str] = None,
                         timeout_in_minutes: Optional[int] = 5) -> Optional[bool]:
        """
        Convenience method to both trigger and poll (wait) on circuit breaker rule execution.

        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute.
        :param namespace: namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :param timeout_in_minutes: Polling timeout in minutes. See poll() for details.
        :return: True if rule execution has breach; False otherwise. See poll() for any exceptions raised.
        """
        breaches = self.poll_all(
                job_execution_uuids=self.trigger_all(rule_uuid=rule_uuid, namespace=namespace, rule_name=rule_name),
                timeout_in_minutes=timeout_in_minutes
        )
        return breaches > 0

    def trigger(self, rule_uuid: Optional[Union[str, UUID]] = None, namespace: Optional[str] = None,
                rule_name: Optional[str] = None) -> str:
        """
        Trigger a rule to start execution with circuit breaker checkpointing.

        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute.
        :param namespace: namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :return: Job execution UUID, as a string, to be used to retrieve execution state / status.
        """
        mutation = Mutation()

        if rule_uuid:
            mutation.trigger_circuit_breaker_rule(rule_uuid=str(rule_uuid)).__fields__('job_execution_uuid')
        elif rule_name:
            if namespace:
                mutation.trigger_circuit_breaker_rule(namespace=namespace, rule_name=rule_name) \
                    .__fields__('job_execution_uuid')
            else:
                mutation.trigger_circuit_breaker_rule(rule_name=rule_name).__fields__('job_execution_uuid')
        else:
            raise ValueError('rule UUID or namespace and rule name must be specified')

        job_execution_uuid = self._client(mutation).trigger_circuit_breaker_rule.job_execution_uuid
        self._print_func(f'Triggered rule with ID \'{rule_uuid}\'. Received \'{job_execution_uuid}\' as execution ID.')
        return job_execution_uuid

    def trigger_all(self, rule_uuid: Optional[Union[str, UUID]] = None, namespace: Optional[str] = None,
                rule_name: Optional[str] = None) -> List[str]:
        """
        Trigger a rule to start execution with circuit breaker checkpointing.

        This function supports rules that create multiple executions (e.g. rules with variables or over
        multiple tables)

        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute.
        :param namespace: namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :return: Job execution UUIDs, as strings, to be used to retrieve execution state / status.
        """
        mutation = Mutation()

        if rule_uuid:
            mutation.trigger_circuit_breaker_rule_v2(rule_uuid=str(rule_uuid)).__fields__('job_execution_uuids')
        elif rule_name:
            if namespace:
                mutation.trigger_circuit_breaker_rule_v2(namespace=namespace, rule_name=rule_name) \
                    .__fields__('job_execution_uuids')
            else:
                mutation.trigger_circuit_breaker_rule_v2(rule_name=rule_name).__fields__('job_execution_uuids')
        else:
            raise ValueError('rule UUID or namespace and rule name must be specified')

        job_execution_uuids = [str(id) for id in self._client(mutation).trigger_circuit_breaker_rule_v2.job_execution_uuids]
        self._print_func(f'Triggered rule with ID \'{rule_uuid}\'. Received {job_execution_uuids} as execution IDs.')
        return job_execution_uuids

    def poll(self, job_execution_uuid: Union[str, UUID], timeout_in_minutes: Optional[int] = 5) -> Optional[int]:
        """
        Poll status / state of an execution for a triggered rule. Polls until status is in a term state or timeout.

        :param job_execution_uuid: UUID for the job execution of a rule (custom SQL monitor).
        :param timeout_in_minutes: Polling timeout in minutes. Note that The Data Collector Lambda has a max timeout of
        15 minutes when executing a query. Queries that take longer to execute are not supported, so we recommend
        filtering down the query output to improve performance (e.g limit WHERE clause). If you expect a query to
        take the full 15 minutes we recommend padding the timeout to 20 minutes.
        :return: Breach count across all executions. A greater than 0 value indicates a breach.
        :raise CircuitBreakerPipelineException: An error in executing the rule (e.g. error in query).
        :raise CircuitBreakerPollException: A timeout during polling or a malformed response.
        """
        return self.poll_all([job_execution_uuid], timeout_in_minutes=timeout_in_minutes)

    def poll_all(self, job_execution_uuids: List[Union[str, UUID]], timeout_in_minutes: Optional[int] = 5) -> Optional[int]:
        """
        Poll status / state of executions for a triggered rule. Polls until status is in a term state or timeout.

        :param job_execution_uuids: UUIDs for the job executions of a rule (custom SQL monitor).
        :param timeout_in_minutes: Polling timeout in minutes. Note that The Data Collector Lambda has a max timeout of
        15 minutes when executing a query. Queries that take longer to execute are not supported, so we recommend
        filtering down the query output to improve performance (e.g limit WHERE clause). If you expect a query to
        take the full 15 minutes we recommend padding the timeout to 20 minutes.
        :return: Breach count. A greater than 0 value indicates a breach.
        :raise CircuitBreakerPipelineException: An error in executing the rule (e.g. error in query).
        :raise CircuitBreakerPollException: A timeout during polling or a malformed response.
        """
        logs = self._poll(job_execution_uuids=job_execution_uuids, timeout_in_minutes=timeout_in_minutes)
        self._print_func(f'Completed polling. Retrieved execution with logs {list(map(str, logs))} for IDs {job_execution_uuids}.')

        breaches = 0
        has_breaches = False
        if logs and len(logs) > 0:
            for log in logs:
                if log.payload.error:
                    logs_str = '\n'.join(str(log) for log in logs)
                    raise CircuitBreakerPipelineException(f'Execution pipeline errored out. Details:\n{logs_str}')
                if log.payload.breach_count is not None:
                    breaches += log.payload.breach_count
                    has_breaches = True

        if has_breaches:
            return breaches
        raise CircuitBreakerPollException

    @boxify(use_snakes=True, default_box_attr=None, default_box=True)
    def _poll(self,
              job_execution_uuids: List[Union[str, UUID]],
              timeout_in_minutes: int,
              sleep_interval_in_seconds: Optional[int] = 15) -> Optional[List[Box]]:
        timeout_start = time.time()
        while time.time() < timeout_start + 60 * timeout_in_minutes:
            query = Query()
            query.get_circuit_breaker_rule_state_v2(job_execution_uuids=map(str, job_execution_uuids)).__fields__('status', 'log')
            circuit_rule_breaker_states = self._client(query).get_circuit_breaker_rule_state_v2

            aggregated_status = self._get_aggregated_status(circuit_rule_breaker_states)
            self._print_func(f'Retrieved execution with aggregated status \'{aggregated_status}\' for '
                             f'IDs {job_execution_uuids}.')

            if aggregated_status in self._TERM_STATES:
                return self._get_payloads(circuit_rule_breaker_states, aggregated_status)

            self._print_func(f'Aggregated state is not terminal state for IDs {job_execution_uuids}. Polling again '
                             f'in \'{sleep_interval_in_seconds}\' seconds.')
            time.sleep(sleep_interval_in_seconds)

    def _get_log_payload(self, log: str):
        log_entries = json.loads(log)
        log_entries.reverse()
        for entry in log_entries:
            if 'payload' in entry:
                return Box(entry, default_box_attr=None, default_box=True)
        return Box()

    def _get_payloads(self, states: List[CircuitBreakerState], status: SqlJobCheckpointStatus) -> List[Box]:
        payloads = []
        for state in states:
            if state.status == status:
                payloads.append(self._get_log_payload(state.log))
        return payloads

    @staticmethod
    def _get_aggregated_status(states: List[CircuitBreakerState]) -> SqlJobCheckpointStatus:
        if not states:
            return SqlJobCheckpointStatus.REGISTERED

        status_by_state = {}
        for state in states:
            status_by_state.setdefault(state.status, []).append(state)

        def all_in_state(s: SqlJobCheckpointStatus):
            return len(status_by_state.get(s, [])) == len(states)

        return (
            SqlJobCheckpointStatus.HAS_ERROR
            if status_by_state.get(SqlJobCheckpointStatus.HAS_ERROR)
            else SqlJobCheckpointStatus.PROCESSING_COMPLETE
            if all_in_state(SqlJobCheckpointStatus.PROCESSING_COMPLETE)
            else SqlJobCheckpointStatus.PROCESSING_START
            if status_by_state.get(SqlJobCheckpointStatus.PROCESSING_COMPLETE) or all_in_state(SqlJobCheckpointStatus.PROCESSING_START)
            else SqlJobCheckpointStatus.EXECUTING_COMPLETE
            if all_in_state(SqlJobCheckpointStatus.PROCESSING_COMPLETE)
            else SqlJobCheckpointStatus.EXECUTING_START
            if status_by_state.get(SqlJobCheckpointStatus.EXECUTING_COMPLETE) or all_in_state(SqlJobCheckpointStatus.EXECUTING_START)
            else SqlJobCheckpointStatus.REGISTERED
        )



from typing import Optional, Union
from uuid import UUID

from pycarlo.core import Client
from pycarlo.features.user.exceptions import MultipleResourcesFoundException, ResourceNotFoundException
from pycarlo.features.user.models import Resource
from pycarlo.features.user.queries import GET_USER_WAREHOUSES


class UserService:

    def __init__(self, mc_client: Optional[Client] = None):
        self._mc_client = mc_client or Client()

    def get_resource(self, resource_id: Optional[Union[str, UUID]] = None) -> Resource:
        """
        Get a resource (e.g. lake or warehouse).

        :param resource_id: resource identifier. If not provided, and your account only has one resource, it will be
                            returned. If your account has multiple resources an exception will be raised indicating a
                            a resource id must be provided.

        :return: resource (e.g. lake or warehouse monitored by Monte Carlo)
        :raise MultipleResourcesFoundException: multiple resources exist (a `resource_id` must be provided)
        :raise ResourceNotFoundException: a resource could not be found
        """
        response = self._mc_client(query=GET_USER_WAREHOUSES)
        warehouses = response.get_user.account.warehouses

        # if resource id was provided, look for matching warehouse record
        if resource_id:
            for w in warehouses:
                if w.uuid == str(resource_id):
                    return Resource(
                        id=UUID(w.uuid),
                        name=w.name,
                        type=w.connection_type
                    )

            # resource not found
            raise ResourceNotFoundException(f'Resource not found with id={resource_id}')

        # if only one warehouse exists, return it
        if len(warehouses) == 1:
            return Resource(
                id=UUID(warehouses[0].uuid),
                name=warehouses[0].name,
                type=warehouses[0].connection_type
            )
        # otherwise, raise an error requesting a resource id
        else:
            raise MultipleResourcesFoundException('Multiple resources found, please specify a resource id')



from unittest import TestCase
from unittest.mock import Mock
from uuid import uuid4, UUID

from box import Box

from pycarlo.features.exceptions import MultipleResourcesFoundException, ResourceNotFoundException
from pycarlo.features.user import UserService
from pycarlo.features.user.queries import GET_USER_WAREHOUSES


class UserServiceTests(TestCase):

    single_resource_response = Box({
        'get_user': {
            'account': {
                'warehouses': [{
                    'uuid': str(uuid4()),
                    'name': 'Snowflake',
                    'connection_type': 'snowflake'
                }]
            }
        }
    })

    multiple_resource_response = Box({
        'get_user': {
            'account': {
                'warehouses': [{
                    'uuid': str(uuid4()),
                    'name': 'Data Lake',
                    'connection_type': 'data-lake'
                }, {
                    'uuid': str(uuid4()),
                    'name': 'Snowflake',
                    'connection_type': 'snowflake'
                }]
            }
        }
    })

    def test_get_default_resource(self):
        # given
        mock_client = Mock(return_value=self.single_resource_response)
        service = UserService(mc_client=mock_client)
        warehouse = self.single_resource_response.get_user.account.warehouses[0]

        # when
        resource = service.get_resource()

        # then
        mock_client.assert_called_once_with(query=GET_USER_WAREHOUSES)
        self.assertEqual(UUID(warehouse.uuid), resource.id)
        self.assertEqual(warehouse.name, resource.name)
        self.assertEqual(warehouse.connection_type, resource.type)

    def test_get_default_resource_with_multiple_resources(self):
        # given
        mock_client = Mock(return_value=self.multiple_resource_response)
        service = UserService(mc_client=mock_client)

        # when
        with self.assertRaises(MultipleResourcesFoundException) as context:
            service.get_resource()

        # then
        mock_client.assert_called_once_with(query=GET_USER_WAREHOUSES)
        self.assertEqual('Multiple resources found, please specify a resource id', str(context.exception))

    def test_get_resource_with_invalid_id(self):
        # given
        mock_client = Mock(return_value=self.multiple_resource_response)
        service = UserService(mc_client=mock_client)
        resource_id = uuid4()

        # when
        with self.assertRaises(ResourceNotFoundException) as context:
            service.get_resource(resource_id)

        # then
        mock_client.assert_called_once_with(query=GET_USER_WAREHOUSES)
        self.assertEqual(f'Resource not found with id={resource_id}', str(context.exception))

    def test_get_resource_with_valid_id(self):
        # given
        mock_client = Mock(return_value=self.multiple_resource_response)
        service = UserService(mc_client=mock_client)
        warehouse = self.multiple_resource_response.get_user.account.warehouses[0]

        # when
        resource = service.get_resource(warehouse.uuid)

        # then
        mock_client.assert_called_once_with(query=GET_USER_WAREHOUSES)
        self.assertEqual(UUID(warehouse.uuid), resource.id)
        self.assertEqual(warehouse.name, resource.name)
        self.assertEqual(warehouse.connection_type, resource.type)



import json
from pathlib import Path
from typing import Callable, Dict, Generic, TypeVar, Union

# file reader return type
T = TypeVar('T')


class FileReader(Generic[T]):
    """
    Utility for reading a local file. Return type is determined by the given `decoder` function.
    """
    def __init__(self, path: Union[Path, str], decoder: Callable[[bytes], T]):
        """
        :param path: local file path
        :param decoder: function that translates file content as bytes into an object of type T
        """
        self._path = to_path(path)
        self._decoder = decoder

    def read(self) -> T:
        """
        Read local file.

        :return: contents of file, represented as type T
        """
        with open(self._path, 'rb') as file:
            content = file.read()
            return self._decoder(content)


class BytesFileReader(FileReader[bytes]):
    """
    Utility for reading a local file as bytes.
    """
    def __init__(self, path: Union[Path, str]):
        """
        :param path: local file path
        """
        super().__init__(
            path=path,
            decoder=lambda b: b
        )


class JsonFileReader(FileReader[Dict]):
    """
    Utility for reading a local JSON file as a dictionary.
    """
    def __init__(self, path: Union[Path, str]):
        """
        :param path: local file path
        """
        super().__init__(
            path=path,
            decoder=lambda b: json.loads(b)
        )


class TextFileReader(FileReader[str]):
    """
    Utility for reading a local file as a string.
    """
    def __init__(self, path: Union[Path, str], encoding: str = 'utf-8'):
        """
        :param path: local file path
        :param encoding: character encoding to use when translating file content to a string
        """
        super().__init__(
            path=path,
            decoder=lambda b: b.decode(encoding)
        )


def to_path(path: Union[Path, str]) -> Path:
    """
    Simple function to normalize a file path passed as either a string or pathlib.Path to a pathlib.Path instance.

    :param path: local file path
    :return: local file path represented as an instance of pathlib.Path
    """
    return path if isinstance(path, Path) else Path(path)



import os
from pathlib import Path

"""Environmental configuration"""

# Enable error logging.
MCD_VERBOSE_ERRORS = os.getenv('MCD_VERBOSE_ERRORS', False) in (True, 'true', 'True')

# MCD API endpoint.
MCD_API_ENDPOINT = os.getenv('MCD_API_ENDPOINT')

# Override MCD Default Profile when reading from the config-file in a session.
MCD_DEFAULT_PROFILE = os.getenv('MCD_DEFAULT_PROFILE')

# Override MCD API ID when creating a session.
MCD_DEFAULT_API_ID = os.getenv('MCD_DEFAULT_API_ID')

# Override MCD API Token when creating a session.
MCD_DEFAULT_API_TOKEN = os.getenv('MCD_DEFAULT_API_TOKEN')

# MCD ID header (for use in local development and testing)
MCD_USER_ID_HEADER = os.getenv('MCD_USER_ID_HEADER')

# dbt cloud API token
DBT_CLOUD_API_TOKEN = os.getenv('DBT_CLOUD_API_TOKEN')

# dbt cloud account ID
DBT_CLOUD_ACCOUNT_ID = os.getenv('DBT_CLOUD_ACCOUNT_ID')

"""Internal Use"""

# Default API endpoint when not provided through env variable nor profile
DEFAULT_MCD_API_ENDPOINT = 'https://api.getmontecarlo.com/graphql'

# Name of the current package.
DEFAULT_PACKAGE_NAME = 'pycarlo'

# Default config keys for the MC config file. Created via the CLI.
DEFAULT_MCD_API_ID_CONFIG_KEY = 'mcd_id'
DEFAULT_MCD_API_TOKEN_CONFIG_KEY = 'mcd_token'
DEFAULT_MCD_API_ENDPOINT_CONFIG_KEY = 'mcd_api_endpoint'

# Default headers for the MC API.
DEFAULT_MCD_API_ID_HEADER = f'x-{DEFAULT_MCD_API_ID_CONFIG_KEY.replace("_", "-")}'
DEFAULT_MCD_API_TOKEN_HEADER = f'x-{DEFAULT_MCD_API_TOKEN_CONFIG_KEY.replace("_", "-")}'
DEFAULT_MCD_USER_ID_HEADER = 'user-id'

# Default headers to trace and help identify requests. For debugging.
DEFAULT_MCD_SESSION_ID = 'x-mcd-session-id'  # Generally the session name.
DEFAULT_MCD_TRACE_ID = 'x-mcd-trace-id'

# File name for profile configuration.
PROFILE_FILE_NAME = 'profiles.ini'

# Default profile to be used.
DEFAULT_PROFILE_NAME = 'default'

# Default path where any configuration files are written.
DEFAULT_CONFIG_PATH = os.path.join(str(Path.home()), '.mcd')

# Default initial wait time for retries in seconds.
DEFAULT_RETRY_INITIAL_WAIT_TIME = 0.25

# Default maximum wait time for retries in seconds.
DEFAULT_RETRY_MAX_WAIT_TIME = 10.0



from utils.vars import ensure_token_env_vars, ensure_endpoint_env_var
from pycarlo.core import Session, Client, Query, Mutation

CONFIRMATION_ICON = '✔️'
CONFIRMATION_SEPARATOR = '='


def test_with_query(client: Client):
    query = Query()
    query.get_user.__fields__('email')
    print(client(query).get_user.email)


def test_with_mutation(client: Client):
    mutation = Mutation()
    mutation.generate_collector_template(region='us-east-1')
    print(client(mutation))


def test_with_string(client: Client):
    get_table_query = """
    query getTables{
      getTables(first: 10) {
        edges {
          node {
            fullTableId
          }
        }
      }
    }
    """
    response = client(get_table_query)
    for edge in response.get_tables.edges:
        print(edge.node.full_table_id)
    print(response['get_tables']['edges'][0]['node']['full_table_id'])


def print_confirmation(text: str):
    separator = CONFIRMATION_SEPARATOR * (len(text) + 2)
    print()
    print(separator)
    print(text, CONFIRMATION_ICON)
    print(separator)
    print()


def sanity_check(mcd_url: str = None, mcd_id: str = None, mcd_token: str = None):
    session = Session(endpoint=mcd_url, mcd_id=mcd_id, mcd_token=mcd_token)
    client = Client(session=session)

    test_with_query(client=client)
    print_confirmation("running queries with Query object")

    test_with_mutation(client=client)
    print_confirmation("running queries with Mutation object")

    test_with_string(client=client)
    print_confirmation("running queries with string")


if __name__ == '__main__':
    api_id, api_token = ensure_token_env_vars()
    api_url = ensure_endpoint_env_var()
    sanity_check(mcd_url=api_url, mcd_id=api_id, mcd_token=api_token)



import json
import os
from datetime import date
from unittest import TestCase

from pycarlo.features.pii import PiiFilterer


class PiiFilteringTests(TestCase):
    FILTERS_1 = {
            'active': [
                {
                    'name': 'a',
                    'pattern': 'n[am]+e',
                },
                {
                    'name': 'b',
                    'pattern': 'su[ab]',
                },
                {
                    'name': 'c',
                    'pattern': 'ag',
                },
                {
                    'name': 'us_ssn',
                    'pattern': r'([0-6]\d{2}|7[0-6]\d|77[0-2])([ \-]?)(\d{2})\2(\d{4})'
                },
                {
                    'name': 'ccard',
                    'pattern': r'((4\d{3})|(5[1-5]\d{2}))(-?|\040?)(\d{4}(-?|\040?)){3}|(3[4,7]\d{2})(-?|\040?)'
                               r'\d{6}(-?|\040?)\d{5}'
                },
                {
                    'name': 'email',
                    'pattern': r'[\w.%+-]+@[\w.-]+\.[\w]{2,4}'
                },
            ]
        }
    FILTERS_2 = {
            'active': [
                {
                    'name': 'us_ssn',
                    'pattern': r'\d{3}-\d{2}-\d{4}'
                },
                {
                    'name': 'ccard',
                    'pattern': r'((4\d{3})|(5[1-5]\d{2}))(-?|\040?)(\d{4}(-?|\040?)){3}|(3[4,7]\d{2})(-?|\040?)'
                               r'\d{6}(-?|\040?)\d{5}'
                },
                {
                    'name': 'email',
                    'pattern': r'[\w.%+-]+@[\w.-]+\.[\w]{2,4}'
                },
                {
                    'name': 'test_col',
                    'pattern': r'COL_\d{5}_\d{2}'
                },
            ]
        }
    FILTERS_COUNTRY_CODE = {
        'active': [
            {
                'name': 'country_code',
                'pattern': '[A-Z]{3}'
            },
        ]
    }

    JSON_MSG_SIMPLE = {
            'a': 'This is my Name: my name',
            'b': [
                'age: 38',
                'and my personal email address is: user_id_09@server.com',
                'This IS a valid SSN: 111-11-1111.',
                'Please charge it to VISA 4321-9876-9876-9876.'
            ],
            'c': 'this is another email address: user-id@server.com',
            'd': 'This IS a valid SSN: 111-11-1111.',
            'e': 'Please charge it to VISA 4321-1234-9876-9876.',
            'f': date(2022, 10, 25),
            'age': 50
        }
    JSON_MSG_STR_LIST = [
        'my personal email address is: user_id_09@server.com',
        'This IS a valid SSN: 111-11-1111.'
    ]
    JSON_MSG_OBJ_LIST = [
        {
            'c': 'this is another email address: user-id@server.com',
            'd': 'This IS a valid SSN: 111-11-1111.'
        },
        {
            'c': 'my personal email address is: user_id_09@server.com',
            'e': 'Please charge it to VISA 4321-1234-9876-9876.',
            'f': date(2022, 10, 25)
        }
    ]
    JSON_MSG_NESTED = {
            'values': [
                [
                    {
                        'x': 'age: 38',
                        'y': 'and my personal email address is: user_id_09@server.com'
                    },
                    {
                        'c': 'this is another email address: user-id@server.com',
                        'f': date(2022, 10, 25),
                        'age': 50
                    }
                ],
                [
                    {
                        'd': 'This IS a valid SSN: 111-11-1111.',
                        'e': 'Please charge it to VISA 4321-1234-9876-9876.',
                    }
                ]
            ],
            'more_values': [
                {
                    'nested_values': [
                        'my personal email address is: user_id_09@server.com',
                        'This IS a valid SSN: 111-11-1111.'
                    ]
                }
            ]
        }

    def test_simple_object_filtering(self):
        f = PiiFilterer(self.FILTERS_1)

        filtered = f.filter_message(self.JSON_MSG_SIMPLE)
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.', filtered['d'])
        self.assertEqual('Please charge it to VISA <filtered:ccard>.', filtered['e'])
        self.assertEqual(self.JSON_MSG_SIMPLE['f'], filtered['f'])
        self.assertEqual(self.JSON_MSG_SIMPLE['age'], filtered['<filtered:c>e'])
        self.assertEqual('<filtered:c>e: 38', filtered['b'][0])
        self.assertEqual('and my personal email address is: <filtered:email>', filtered['b'][1])
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.', filtered['b'][2])
        self.assertEqual('Please charge it to VISA <filtered:ccard>.', filtered['b'][3])

        m = filtered['pii_metrics']  # equivalent to f.metrics
        self.assertEqual(1, m['a']['replacements'])
        self.assertEqual(0, m['b']['replacements'])
        self.assertEqual(2, m['c']['replacements'])
        self.assertEqual(2, m['us_ssn']['replacements'])
        self.assertEqual(2, m['email']['replacements'])
        self.assertEqual(9, m['_total']['replacements'])
        self.assertGreater(m['_total']['time_taken_ms'], 0)

    def test_string_list_filtering(self):
        f = PiiFilterer(self.FILTERS_1)

        filtered = f.filter_message(self.JSON_MSG_STR_LIST)
        self.assertEqual('my personal email address is: <filtered:email>', filtered[0])
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.', filtered[1])

    def test_object_list_filtering(self):
        f = PiiFilterer(self.FILTERS_1)

        filtered = f.filter_message(self.JSON_MSG_OBJ_LIST)
        self.assertEqual('this is another email address: <filtered:email>', filtered[0]['c'])
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.', filtered[0]['d'])
        self.assertEqual('my personal email address is: <filtered:email>', filtered[1]['c'])
        self.assertEqual('Please charge it to VISA <filtered:ccard>.', filtered[1]['e'])
        self.assertEqual(self.JSON_MSG_OBJ_LIST[1]['f'], filtered[1]['f'])

    def test_nested_object_filtering(self):
        f = PiiFilterer(self.FILTERS_1)

        filtered = f.filter_message(self.JSON_MSG_NESTED)
        self.assertEqual('<filtered:c>e: 38', filtered['values'][0][0]['x'])
        self.assertEqual('and my personal email address is: <filtered:email>', filtered['values'][0][0]['y'])
        self.assertEqual('this is another email address: <filtered:email>', filtered['values'][0][1]['c'])
        self.assertEqual(self.JSON_MSG_NESTED['values'][0][1]['f'], filtered['values'][0][1]['f'])
        self.assertEqual(self.JSON_MSG_NESTED['values'][0][1]['age'], filtered['values'][0][1]['<filtered:c>e'])
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.', filtered['values'][1][0]['d'])
        self.assertEqual('Please charge it to VISA <filtered:ccard>.', filtered['values'][1][0]['e'])

        self.assertEqual('my personal email address is: <filtered:email>',
                         filtered['more_values'][0]['nested_values'][0])
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.',
                         filtered['more_values'][0]['nested_values'][1])

        m = filtered['pii_metrics']
        self.assertEqual(2, m['us_ssn']['replacements'])
        self.assertEqual(3, m['email']['replacements'])
        self.assertEqual(1, m['ccard']['replacements'])
        self.assertEqual(2, m['c']['replacements'])
        self.assertEqual(8, m['_total']['replacements'])

    def test_filter_md_events(self):
        f = PiiFilterer(self.FILTERS_2)

        dir_name = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_events')
        event_files = os.listdir(dir_name)
        email_count = 0
        for event_file in event_files:
            if not event_file.endswith('.json'):
                continue
            with open(os.path.join(dir_name, event_file), 'r') as file:
                data = json.load(file)
                filtered_data = f.filter_message(data)
                metrics = filtered_data['pii_metrics']
                email_count += metrics.get('email')['replacements']

        self.assertGreaterEqual(email_count, 1)  # there's at least one email in sample_query_log_01.json

    def test_filter_tuples_single_column(self):
        data = {'rows': [("4",), ("3",), ("USA",)]}
        f = PiiFilterer(self.FILTERS_COUNTRY_CODE)
        filtered_data = f.filter_message(data)

        self.assertEqual('4', filtered_data['rows'][0][0])
        self.assertEqual('3', filtered_data['rows'][1][0])
        self.assertEqual('<filtered:country_code>', filtered_data['rows'][2][0])

    def test_filter_tuples_three_columns(self):
        data = {'rows': [
            (date(2022, 12, 1), "4", 23),
            (date(2022, 12, 2), "3", 15),
            (date(2022, 12, 3), "USA", 35)]}
        f = PiiFilterer(self.FILTERS_COUNTRY_CODE)
        filtered_data = f.filter_message(data)

        self.assertEqual(data['rows'][0][0], filtered_data['rows'][0][0])
        self.assertEqual(data['rows'][0][1], filtered_data['rows'][0][1])
        self.assertEqual(data['rows'][0][2], filtered_data['rows'][0][2])

        self.assertEqual(data['rows'][2][0], filtered_data['rows'][2][0])
        self.assertEqual('<filtered:country_code>', filtered_data['rows'][2][1])
        self.assertEqual(data['rows'][2][2], filtered_data['rows'][2][2])

    def test_filter_bytes_utf(self):
        data = '[\u00f1\u00f1] This IS a valid SSN: 111-11-1111.'.encode('utf8')
        f = PiiFilterer(self.FILTERS_2)
        filtered_data = f.filter_data(data)  # default encoding is utf8
        filtered_text = filtered_data.decode('utf8')
        self.assertEqual('[\u00f1\u00f1] This IS a valid SSN: <filtered:us_ssn>.', filtered_text)

    def test_filter_str(self):
        text = 'This IS a valid SSN: 111-11-1111.'
        f = PiiFilterer(self.FILTERS_2)
        filtered_text = f.filter_str(text)
        self.assertEqual('This IS a valid SSN: <filtered:us_ssn>.', filtered_text)

    def test_no_filter_bytes(self):
        text = '[\u00f1\u00f1] This IS a valid SSN: 111-11-1111.'
        data = text.encode('utf8')
        f = PiiFilterer(filters_config=None)
        filtered_data = f.filter_data(data)  # default encoding is utf8
        filtered_text = filtered_data.decode('utf8')
        self.assertEqual(text, filtered_text)

    def test_no_filter_str(self):
        text = 'This IS a valid SSN: 111-11-1111.'
        f = PiiFilterer(filters_config=None)
        filtered_text = f.filter_str(text)
        self.assertEqual(text, filtered_text)

    def test_no_filter_object(self):
        f = PiiFilterer(filters_config=None)

        filtered = f.filter_message(self.JSON_MSG_SIMPLE)
        self.assertEqual(filtered, self.JSON_MSG_SIMPLE)

    def test_filter_bytes_ascii(self):
        enc = 'iso-8859-1'
        data = '[\u00f1\u00f1] This IS a valid SSN: 111-11-1111.'.encode(enc)
        f = PiiFilterer(self.FILTERS_2)
        filtered_data = f.filter_data(data, enc)
        filtered_text = filtered_data.decode(enc)
        self.assertEqual('[\u00f1\u00f1] This IS a valid SSN: <filtered:us_ssn>.', filtered_text)
